"""
Helpers for writing clients.
"""

import asyncio
import httpx
import mcp
import mcp.types
from contextlib import asynccontextmanager
from mcp.client.session import ElicitationFnT, MessageHandlerFnT
from mcp.client.streamable_http import streamablehttp_client
from mcp.server.streamable_http import (
    MCP_PROTOCOL_VERSION_HEADER,
    MCP_SESSION_ID_HEADER,
)
from typing import AsyncIterator, Any, Callable


def create_mcp_http_client(
    headers: dict[str, str] | None = None,
    timeout: httpx.Timeout | None = None,
    auth: httpx.Auth | None = None,
) -> httpx.AsyncClient:
    """
    Almost identical implementation to
    `mcp.shared._httpx_utils.create_mcp_http_client`, but we've added
    a transport which does retries.
    """
    kwargs: dict[str, Any] = {
        "follow_redirects": True,
    }

    kwargs["timeout"] = timeout or httpx.Timeout(30.0)

    if headers is not None:
        kwargs["headers"] = headers

    if auth is not None:
        kwargs["auth"] = auth

    return httpx.AsyncClient(
        transport=httpx.AsyncHTTPTransport(retries=5),
        **kwargs,
    )


@asynccontextmanager
async def connect(
    url: str,
    *,
    headers: dict[str, str] | None = None,
    terminate_on_close: bool = True,
    elicitation_callback: ElicitationFnT | None = None,
    message_handler: MessageHandlerFnT | None = None,
) -> AsyncIterator[tuple[mcp.ClientSession, str, str | int]]:
    async with streamablehttp_client(
        url,
        headers=headers,
        terminate_on_close=terminate_on_close,
        httpx_client_factory=create_mcp_http_client,
    ) as (read_stream, write_stream, get_session_id):
        async with mcp.ClientSession(
            read_stream,
            write_stream,
            elicitation_callback=elicitation_callback,
            message_handler=message_handler,
        ) as session:
            result = await session.initialize()
            assert isinstance(result, mcp.types.InitializeResult)

            protocol_version = result.protocolVersion

            session_id = get_session_id()

            assert session_id is not None

            yield session, session_id, protocol_version


@asynccontextmanager
async def reconnect(
    url: str,
    *,
    session_id: str,
    protocol_version: str | int,
    next_request_id: int,
    terminate_on_close: bool = True,
    elicitation_callback: ElicitationFnT | None = None,
    message_handler: MessageHandlerFnT | None = None,
) -> AsyncIterator[mcp.ClientSession]:
    headers: dict[str, Any] = {}
    headers[MCP_SESSION_ID_HEADER] = session_id
    headers[MCP_PROTOCOL_VERSION_HEADER] = protocol_version
    async with streamablehttp_client(
        url,
        headers=headers,
        terminate_on_close=terminate_on_close,
        httpx_client_factory=create_mcp_http_client,
    ) as (read_stream, write_stream, get_session_id):
        async with mcp.ClientSession(
            read_stream,
            write_stream,
            elicitation_callback=elicitation_callback,
            message_handler=message_handler,
        ) as session:
            session._request_id = next_request_id
            yield session



import mcp.types
import pickle
from contextvars import ContextVar
from anyio import create_memory_object_stream
from anyio.streams.memory import (
    MemoryObjectReceiveStream,
    MemoryObjectSendStream,
)
from contextlib import contextmanager
from dataclasses import dataclass
from log.log import get_logger
from mcp.server.lowlevel.server import Server
from mcp.shared.message import SessionMessage
from rbt.mcp.v1.session_rbt import (
    HandleMessageRequest,
    HandleMessageResponse,
    RunRequest,
    RunResponse,
    Session,
)
from rbt.mcp.v1.stream_rbt import Stream
from reboot.aio.auth.authorizers import allow
from reboot.aio.contexts import WorkflowContext, WriterContext
from reboot.aio.workflows import at_least_once
from reboot.mcp.event_store import get_event_id

logger = get_logger(__name__)

# Dictionary from path to MCP server for calling `run()`.
_servers: dict[str, Server] = {}

# Python asyncio context variable for the `WorkflowContext` of the
# current message being handled.
_context: ContextVar[WorkflowContext | None] = ContextVar(
    "`WorkflowContext` of current message being handled",
    default=None,
)


@dataclass(kw_only=True)
class Streams:
    refs: int
    read_stream: tuple[MemoryObjectSendStream[SessionMessage | Exception],
                       MemoryObjectReceiveStream[SessionMessage | Exception]]
    write_stream: tuple[MemoryObjectSendStream[SessionMessage],
                        MemoryObjectReceiveStream[SessionMessage]]


class SessionServicer(Session.Servicer):

    def __init__(self):
        self._request_streams: dict[mcp.types.RequestId, Streams] = {}

        # For requests coming from the server, e.g., elicit, sample,
        # etc, this maps the event ID that we generated to a tuple of
        # the server generated request ID and the "related request ID"
        # for looking up the stream to send the response.
        self._write_request_ids: dict[
            str,
            # (server generated request ID, related request ID)
            tuple[mcp.types.RequestId, mcp.types.RequestId]
        ] = {}

    def authorizer(self):
        return allow()

    @contextmanager
    def _get_request_streams(self, request_id: mcp.types.RequestId):
        # `mcp.types.RequestId` is of type `str | int`, and it seems
        # the mcp SDK sometimes uses a `str` and sometimes uses an
        # `int` for the _same_ request ID, e.g., `1`, so we always
        # need to canonicalize it as a `str`.
        request_id = str(request_id)
        try:
            if request_id not in self._request_streams:
                # Create streams for communicating with MCP server.
                self._request_streams[request_id] = Streams(
                    refs=1,  # Initial reference count.
                    read_stream=create_memory_object_stream[
                        SessionMessage | Exception](),
                    write_stream=create_memory_object_stream[SessionMessage](),
                )
            else:
                self._request_streams[request_id].refs += 1

            yield (
                self._request_streams[request_id].read_stream,
                self._request_streams[request_id].write_stream,
            )
        finally:
            self._request_streams[request_id].refs -= 1
            if self._request_streams[request_id].refs == 0:
                # TODO: do we also need to close the streams in order
                # for them to get garbage collected?
                del self._request_streams[request_id]

    async def HandleMessage(
        self,
        context: WorkflowContext,
        request: HandleMessageRequest,
    ) -> HandleMessageResponse:
        message = pickle.loads(request.message_bytes)

        if isinstance(message.message.root, mcp.types.JSONRPCRequest):
            logger.debug(f"Handling ({type(message).__name__}): {message}")

            request_id = message.message.root.id

            stream = Stream.ref(str(request_id))

            with self._get_request_streams(
                request_id,
            ) as (read_stream, write_stream):
                read_stream_send, _ = read_stream
                _, write_stream_receive = write_stream

                run_task = await self.ref().spawn().Run(
                    context,
                    path=request.path,
                    message_bytes=request.message_bytes,
                )

                async def send_and_receive():

                    await read_stream_send.send(message)

                    async for write_message in write_stream_receive:
                        logger.debug(
                            f"Sending message ({type(write_message).__name__}): "
                            f"{write_message}"
                        )

                        event_id = get_event_id(write_message)

                        # If this is a request, we need to grab the
                        # request ID and map it to something else that
                        # we actually send so that we can reconnect it
                        # here.
                        if isinstance(write_message.message.root, mcp.types.JSONRPCRequest):
                            write_request_id = write_message.message.root.id
                            write_message.message.root.id = event_id
                            related_request_id = write_message.metadata.related_request_id
                            assert related_request_id is not None
                            self._write_request_ids[event_id] = (write_request_id, related_request_id)

                        await stream.per_workflow(event_id).Put(
                            context,
                            event_id=event_id,
                            message_bytes=pickle.dumps(write_message),
                        )

                        if isinstance(
                            write_message.message.root,
                            mcp.types.JSONRPCResponse | mcp.types.JSONRPCError,
                        ):
                            break

                await at_least_once(
                    "Send and receive",
                    context,
                    send_and_receive,
                )

                await read_stream_send.aclose()

                await run_task

                logger.debug(f"Completed ({type(message).__name__}): {message}")

                return HandleMessageResponse()

        elif isinstance(message.message.root, mcp.types.JSONRPCNotification):
            # Ignore "notifications/initialized" as we run the
            # servers with `stateless=True` here so they are
            # always initialized.
            if message.message.root.method == "notifications/initialized":
                return HandleMessageResponse()

            # TODO: handle notification or route to the
            # appropriate request stream if relevant.

        elif isinstance(message.message.root, mcp.types.JSONRPCResponse):
            # We override outgoing request IDs with our own event IDs
            # to make them unique and routable, and they are always
            # strings so we ensure that here.
            event_id = str(message.message.root.id)

            request_id, related_request_id = self._write_request_ids[event_id]

            message.message.root.id = request_id

            with self._get_request_streams(
                related_request_id,
            ) as (read_stream, _):
                read_stream_send, _ = read_stream

                await read_stream_send.send(message)

            return HandleMessageResponse()

        logger.warning(f"UNIMPLEMENTED ({type(message).__name__}): {message}")

        return HandleMessageResponse()

    async def Run(
        self,
        context: WorkflowContext,
        request: RunRequest,
    ) -> RunResponse:
        path = request.path
        message = pickle.loads(request.message_bytes)

        assert isinstance(message.message.root, mcp.types.JSONRPCRequest)

        request_id = message.message.root.id

        with self._get_request_streams(
            request_id,
        ) as (read_stream, write_stream):
            _, read_stream_receive = read_stream
            write_stream_send, _ = write_stream

            async def server_run():
                global _servers
                server = _servers[path]
                _context.set(context)
                try:
                    await server.run(
                        read_stream_receive,
                        write_stream_send,
                        server.create_initialization_options(),
                        raise_exceptions=True,
                        # Since we might resume we set `stateless=True`
                        # because we don't want the server to need to do
                        # initialization, but it will happily do it when the
                        # client does it on connect.
                        stateless=True,
                    )
                except:
                    import traceback
                    traceback.print_exc()
                    raise
                finally:
                    _context.set(None)

            await at_least_once("Server run", context, server_run)

            return RunResponse()



import mcp.types
import pickle
from mcp.server.streamable_http import (
    GET_STREAM_KEY,
    EventCallback,
    EventId,
    EventMessage,
    StreamId,
)
from mcp.server.streamable_http import EventStore
from mcp.shared.message import ServerMessageMetadata, SessionMessage
from rbt.mcp.v1.stream_rbt import Stream
from reboot.aio.external import ExternalContext
from uuid import uuid4


def get_stream_id(message: SessionMessage) -> StreamId:
    # Try and extract the ID of the original request.
    request_id = None

    if isinstance(
        message.message.root,
        mcp.types.JSONRPCResponse | mcp.types.JSONRPCError,
    ):
        request_id = str(message.message.root.id)
    elif (
        message.metadata is not None and
        isinstance(message.metadata, ServerMessageMetadata) and
        message.metadata.related_request_id is not None
    ):
        request_id = str(message.metadata.related_request_id)

    stream_id = request_id if request_id is not None else GET_STREAM_KEY

    return stream_id


def get_event_id(message: SessionMessage) -> EventId:
    try:
        if isinstance(
            message.message.root,
            mcp.types.JSONRPCRequest | mcp.types.JSONRPCNotification,
        ):
            assert (
                message.message.root.params is not None and
                "_meta" in message.message.root.params and
                "rebootEventId" in message.message.root.params["_meta"]
            ), f"Missing event ID for {message.message.root}"

            return message.message.root.params["_meta"]["rebootEventId"]

        assert isinstance(
            message.message.root,
            mcp.types.JSONRPCResponse | mcp.types.JSONRPCError,
        )

        # This is the original request ID which is sufficient for
        # differentiation.
        return str(message.message.root.id)
    except:
        # TODO: remove once we've support for all types of `message`s
        import traceback
        traceback.print_exc()
        raise


def qualified_event_id(stream_id: StreamId, event_id: EventId) -> EventId:
    return f"{stream_id}/{event_id}"


def stream_id_from_qualified_event_id(event_id: EventId) -> StreamId:
    index = event_id.rfind("/")
    assert index != -1
    return event_id[:index]


class DurableEventStore(EventStore):

    def __init__(self, context: ExternalContext):
        self._context = context

    async def store_event(
        self,
        stream_id: StreamId,
        message: mcp.types.JSONRPCMessage,
    ) -> EventId:
        event_id = get_event_id(SessionMessage(message=message))
        assert event_id is not None
        return qualified_event_id(stream_id, event_id)

    async def replay_events_after(
        self,
        last_event_id: EventId,
        send_callback: EventCallback,
    ) -> StreamId | None:
        stream_id = stream_id_from_qualified_event_id(last_event_id)

        async for message, event_id in replay(
            self._context,
            stream_id=stream_id,
            last_event_id=last_event_id,
        ):
            await send_callback(EventMessage(message.message, event_id))

        return stream_id


async def replay(
    context: ExternalContext,
    *,
    stream_id: StreamId,
    last_event_id: EventId | None = None,
):
    stream = Stream.ref(stream_id)

    # Ensure the stream has been created.
    await stream.Create(context)

    # TODO: fix `.reactively()` so we don't need the `while True`.
    while True:
        async for replay in stream.reactively().Replay(
            context,
            last_event_id=last_event_id,
        ):
            if len(replay.events) == 0:
                continue

            for event in replay.events:
                message = pickle.loads(event.message_bytes)
                yield message, event.id
                if isinstance(
                    message.message.root,
                    mcp.types.JSONRPCResponse | mcp.types.JSONRPCError,
                ):
                    return

            last_event_id = replay.events[-1].id
            break
                                    




"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class Stream(google.protobuf.message.Message):
    """//////////////////////////////////////////////////////////////////////"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    EVENTS_FIELD_NUMBER: builtins.int
    @property
    def events(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Event]: ...
    def __init__(
        self,
        *,
        events: collections.abc.Iterable[global___Event] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["events", b"events"]) -> None: ...

global___Stream = Stream

@typing_extensions.final
class Event(google.protobuf.message.Message):
    """//////////////////////////////////////////////////////////////////////"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ID_FIELD_NUMBER: builtins.int
    MESSAGE_BYTES_FIELD_NUMBER: builtins.int
    id: builtins.str
    message_bytes: builtins.bytes
    def __init__(
        self,
        *,
        id: builtins.str = ...,
        message_bytes: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["id", b"id", "message_bytes", b"message_bytes"]) -> None: ...

global___Event = Event

@typing_extensions.final
class PutRequest(google.protobuf.message.Message):
    """//////////////////////////////////////////////////////////////////////"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    EVENT_ID_FIELD_NUMBER: builtins.int
    MESSAGE_BYTES_FIELD_NUMBER: builtins.int
    event_id: builtins.str
    message_bytes: builtins.bytes
    def __init__(
        self,
        *,
        event_id: builtins.str = ...,
        message_bytes: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["event_id", b"event_id", "message_bytes", b"message_bytes"]) -> None: ...

global___PutRequest = PutRequest

@typing_extensions.final
class PutResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___PutResponse = PutResponse

@typing_extensions.final
class ReplayRequest(google.protobuf.message.Message):
    """//////////////////////////////////////////////////////////////////////"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    LAST_EVENT_ID_FIELD_NUMBER: builtins.int
    last_event_id: builtins.str
    def __init__(
        self,
        *,
        last_event_id: builtins.str | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_last_event_id", b"_last_event_id", "last_event_id", b"last_event_id"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_last_event_id", b"_last_event_id", "last_event_id", b"last_event_id"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_last_event_id", b"_last_event_id"]) -> typing_extensions.Literal["last_event_id"] | None: ...

global___ReplayRequest = ReplayRequest

@typing_extensions.final
class ReplayResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    EVENTS_FIELD_NUMBER: builtins.int
    @property
    def events(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Event]: ...
    def __init__(
        self,
        *,
        events: collections.abc.Iterable[global___Event] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["events", b"events"]) -> None: ...

global___ReplayResponse = ReplayResponse



from google.protobuf.empty_pb2 import Empty
from rbt.mcp.v1.stream_rbt import (
    Event,
    PutRequest,
    PutResponse,
    ReplayRequest,
    ReplayResponse,
    Stream,
)
from reboot.aio.auth.authorizers import allow
from reboot.aio.contexts import ReaderContext, WriterContext
from reboot.mcp.event_store import qualified_event_id


class StreamServicer(Stream.Servicer):

    def authorizer(self):
        return allow()

    async def Create(
        self,
        context: WriterContext,
        request: Empty,
    ) -> Empty:
        return Empty()

    async def Put(
        self,
        context: WriterContext,
        request: PutRequest,
    ) -> PutResponse:
        self.state.events.append(
            Event(
                id=qualified_event_id(context.state_id, request.event_id),
                message_bytes=request.message_bytes,
            )
        )
        return PutResponse()

    async def Replay(
        self,
        context: ReaderContext,
        request: ReplayRequest,
    ) -> ReplayResponse:
        if not request.HasField("last_event_id"):
            return ReplayResponse(events=self.state.events)

        for i, event in enumerate(self.state.events):
            if event.id == request.last_event_id:
                return ReplayResponse(events=self.state.events[i+1:])

        return ReplayResponse()



import asyncio
import functools
import httpx
import inspect
import logging
import mcp.types
import pickle
from dataclasses import dataclass
from log.log import get_logger, set_log_level
from mcp.server import fastmcp
from mcp.server.elicitation import (
    AcceptedElicitation,
    CancelledElicitation,
    DeclinedElicitation,
    ElicitationResult,
    ElicitSchemaModelT,
    _validate_elicitation_schema,
)
from mcp.server.session import ServerSession
from mcp.server.streamable_http import (
    MCP_SESSION_ID_HEADER,
    StreamableHTTPServerTransport,
)
from mcp.shared.message import ServerMessageMetadata
from rbt.mcp.v1.session_rbt import Session
from reboot.aio.applications import Application
from reboot.aio.contexts import WorkflowContext
from reboot.aio.external import ExternalContext
from reboot.aio.types import StateRef
from reboot.mcp.event_store import DurableEventStore, replay
from reboot.mcp.servicers.session import (
    SessionServicer,
    _servers,
    _context,
)
from reboot.aio.workflows import at_least_once
from reboot.mcp.servicers.stream import StreamServicer
from reboot.std.collections.v1 import sorted_map
from rebootdev.aio.headers import CONSENSUS_ID_HEADER, STATE_REF_HEADER
from rebootdev.memoize.v1.memoize_rbt import Memoize
from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import Response, StreamingResponse
from starlette.routing import Route
from starlette.types import Receive, Scope, Send
from types import MethodType
from typing import Any, Callable, Literal, Protocol, TypeAlias, cast
from uuid import uuid4, uuid5

logger = get_logger(__name__)


class DurableSession:

    _session: ServerSession
    _context: WorkflowContext
    _event_aliases: set[str]

    def __init__(self, session: ServerSession, context: WorkflowContext):
        self._session = session
        self._context = context
        self._event_aliases = set()

    def _event_id(self, function_name: str, why: str):
        event_alias = f"{function_name}: {why}"

        assert self._context is not None

        if self._context.within_loop():
            event_alias += f" #{self._context.task.iteration}"

        if event_alias in self._event_aliases:
            raise TypeError(
                f"Looks like you're calling `{function_name}()` "
                "more than once with the same `why`"
            )

        self._event_aliases.add(event_alias)

        workflow_id = self._context.workflow_id

        assert workflow_id is not None

        # Generate a unique but deterministic ID for this event based
        # on the alias and this workflow (which is unique per
        # request).
        return uuid5(workflow_id, event_alias).hex

    async def send_resource_list_changed(self, why: str) -> None:
        """
        Send a resource list changed notification.

        Args:
            why: Description of why the resource list changed,
                 used to durably differentiate resource list
                 changed events that are sent to client
        """
        event_id = self._event_id("send_resource_list_changed", why)

        await self._session.send_notification(
            mcp.types.ServerNotification(
                mcp.types.ResourceListChangedNotification(
                    # TODO: figure out why `mypy` requires
                    # passing `method` which has a default.
                    method="notifications/resources/list_changed",
                    params=mcp.types.NotificationParams(
                        _meta=mcp.types.NotificationParams.Meta(
                            rebootEventId=str(event_id),
                        ),
                    ),
                ),
            ),
        )

    async def send_tool_list_changed(self, why: str) -> None:
        """
        Send a tool list changed notification.

        Args:
            why: Description of why the resource list changed,
                 used to durably differentiate resource list
                 changed events that are sent to client
        """
        event_id = self._event_id("send_tool_list_changed", why)

        await self._session.send_notification(
            mcp.types.ServerNotification(
                mcp.types.ToolListChangedNotification(
                    # TODO: figure out why `mypy` requires
                    # passing `method` which has a default.
                    method="notifications/tools/list_changed",
                    params=mcp.types.NotificationParams(
                        _meta=mcp.types.NotificationParams.Meta(
                            rebootEventId=str(event_id),
                        ),
                    ),
                ),
            ),
        )

    async def send_prompt_list_changed(self, why: str) -> None:
        """
        Send a prompt list changed notification.

        Args:
            why: Description of why the resource list changed,
                 used to durably differentiate resource list
                 changed events that are sent to client
        """
        event_id = self._event_id("send_prompt_list_changed", why)

        await self._session.send_notification(
            mcp.types.ServerNotification(
                mcp.types.PromptListChangedNotification(
                    # TODO: figure out why `mypy` requires
                    # passing `method` which has a default.
                    method="notifications/prompts/list_changed",
                    params=mcp.types.NotificationParams(
                        _meta=mcp.types.NotificationParams.Meta(
                            rebootEventId=str(event_id),
                        ),
                    ),
                ),
            ),
        )


class DurableContextProtocol(Protocol):

    _event_aliases: set[str]

    session: DurableSession

    async def report_progress(
        self,
        progress: float,
        total: float | None = None,
        message: str | None = None,
    ) -> None:
        """Report progress for the current operation.

        Args:
            progress: Current progress value e.g. 24
            total: Optional total value e.g. 100
            message: Optional message e.g. Starting render...
        """
        ...

    async def log(
        self,
        level: Literal["debug", "info", "warning", "error"],
        message: str,
        *,
        logger_name: str | None = None,
    ) -> None:
        """Send a log message to the client.

        Args:
            level: Log level (debug, info, warning, error)
            message: Log message
            logger_name: Optional logger name
            **data: Additional structured data to include
        """
        ...

    async def debug(self, message: str) -> None:
        """Send a debug log message."""
        ...

    async def info(self, message: str) -> None:
        """Send an info log message."""
        ...

    async def warning(self, message: str) -> None:
        """Send a warning log message."""
        ...

    async def error(self, message: str) -> None:
        """Send an error log message."""
        ...

    async def elicit(
        self,
        message: str,
        schema: type[ElicitSchemaModelT],
    ) -> ElicitationResult:
         """
         Elicit information from the client/user.

        This method can be used to interactively ask for additional
        information from the client within a tool's execution. The
        client might display the message to the user and collect a
        response according to the provided schema. Or in case a client
        is an agent, it might decide how to handle the elicitation --
        either by asking the user or automatically generating a
        response.

        Args:
            schema: A Pydantic model class defining the expected
                    response structure, according to the
                    specification, only primive types are allowed.
            message: Optional message to present to the user. If not
                     provided, will use a default message based on the
                     schema

        Returns:
            An ElicitationResult containing the action taken and the
            data if accepted

        Note:
            Check the result.action to determine if the user accepted,
            declined, or cancelled.  The result.data will only be
            populated if action is "accept" and validation succeeded.
         """
         ...


class DurableContext(WorkflowContext, DurableContextProtocol):

    pass


@dataclass(kw_only=True, frozen=True)
class Resource:
    fn: mcp.types.AnyFunction
    uri: str
    name: str | None
    title: str | None
    description: str | None
    mime_type: str | None


@dataclass(kw_only=True, frozen=True)
class Prompt:
    func: mcp.types.AnyFunction
    name: str | None
    title: str | None
    description: str | None


@dataclass(kw_only=True, frozen=True)
class Tool:
    fn: mcp.types.AnyFunction
    name: str | None
    title: str | None
    description: str | None
    annotations: mcp.types.ToolAnnotations | None
    structured_output: bool | None


LogLevel: TypeAlias = Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]


class DurableMCP:
    """
    Proxy for `fastmcp.FastMCP`, but wrapping tools, prompts, and
    resources appropriately to make them durable.

    NOTE: this class explicitly does NOT extend from `fastmcp.FastMCP`
    so that we don't mislead users into thinking some features are
    implemented that are not.
    """

    _log_level: LogLevel
    _resources: list[Resource]
    _prompts: list[Prompt]
    _tools: list[Tool]

    def __init__(
        self,
        *,
        path: str,
        log_level: LogLevel = "WARNING",
    ):
        self._log_level = log_level
        self._path = path
        self._resources = []
        self._prompts = []
        self._tools = []

        set_log_level(logging.getLevelNamesMapping()[log_level])

    @property
    def path(self):
        return self._path

    def servicers(self):
        return [SessionServicer, StreamServicer] + sorted_map.servicers()

    def resource(
        self,
        uri: str,
        *,
        name: str | None = None,
        title: str | None = None,
        description: str | None = None,
        mime_type: str | None = None,
    ) -> Callable[[mcp.types.AnyFunction], mcp.types.AnyFunction]:
        """Decorator to register a function as a resource.

        The function will be called when the resource is read to generate its content.
        The function can return:
        - str for text content
        - bytes for binary content
        - other types will be converted to JSON

        If the URI contains parameters (e.g. "resource://{param}") or the function
        has parameters, it will be registered as a template resource.

        Args:
            uri: URI for the resource (e.g. "resource://my-resource" or "resource://{param}")
            name: Optional name for the resource
            title: Optional human-readable title for the resource
            description: Optional description of the resource
            mime_type: Optional MIME type for the resource

        Example:
            @server.resource("resource://my-resource")
            def get_data() -> str:
                return "Hello, world!"

            @server.resource("resource://my-resource")
            async get_data() -> str:
                data = await fetch_data()
                return f"Hello, world! {data}"

            @server.resource("resource://{city}/weather")
            def get_weather(city: str) -> str:
                return f"Weather for {city}"

            @server.resource("resource://{city}/weather")
            async def get_weather(city: str) -> str:
                data = await fetch_weather(city)
                return f"Weather for {city}: {data}"
        """
        # Check if user passed function directly instead of calling decorator.
        if callable(uri):
            raise TypeError(
                "The @resource decorator was used incorrectly. "
                "Did you forget to call it? Use @resource('uri') instead of @resource"
            )

        def decorator(fn: mcp.types.AnyFunction) -> mcp.types.AnyFunction:
            self._resources.append(
                Resource(
                    fn=fn,
                    uri=uri,
                    name=name,
                    title=title,
                    description=description,
                    mime_type=mime_type,
                )
            )
            return fn

        return decorator

    def add_resource(self, resource: fastmcp.resources.Resource) -> None:
        # Where as `add_tool()` is a great insertion point,
        # `add_resource` gives us an already modified function which
        # does not pickle to the child process by default. This
        # probably isn't an issue as mostly folks will just use the
        # decorator, but if it is, we can try extracting everything
        # from `resource`? We also have to override the `resource()`
        # decorator anyway to get templates.
        raise NotImplementedError("Use `resource()` decorator instead")


    def prompt(
        self,
        name: str | None = None,
        title: str | None = None,
        description: str | None = None,
    ) -> Callable[[mcp.types.AnyFunction], mcp.types.AnyFunction]:
        """Decorator to register a prompt.

        Args:
            name: Optional name for the prompt (defaults to function name)
            title: Optional human-readable title for the prompt
            description: Optional description of what the prompt does

        Example:
            @server.prompt()
            def analyze_table(table_name: str) -> list[Message]:
                schema = read_table_schema(table_name)
                return [
                    {
                        "role": "user",
                        "content": f"Analyze this schema:\n{schema}"
                    }
                ]

            @server.prompt()
            async def analyze_file(path: str) -> list[Message]:
                content = await read_file(path)
                return [
                    {
                        "role": "user",
                        "content": {
                            "type": "resource",
                            "resource": {
                                "uri": f"file://{path}",
                                "text": content
                            }
                        }
                    }
                ]
        """
        # Check if user passed function directly instead of calling decorator.
        if callable(name):
            raise TypeError(
                "The @prompt decorator was used incorrectly. "
                "Did you forget to call it? Use @prompt() instead of @prompt"
            )

        def decorator(func: mcp.types.AnyFunction) -> mcp.types.AnyFunction:
            self._prompts.append(
                Prompt(
                    func=func,
                    name=name,
                    title=title,
                    description=description,
                ),
            )
            return func

        return decorator

    def add_prompt(self, prompt: fastmcp.prompts.Prompt) -> None:
        # Where as `add_tool()` is a great insertion point,
        # `add_prompt` gives us an already modified function which
        # does not pickle to the child process by default. This
        # probably isn't an issue as mostly folks will just use the
        # decorator, but if it is, we can try extracting everything
        # from `prompt`?
        raise NotImplementedError("Use `prompt()` decorator instead")

    def tool(
        self,
        name: str | None = None,
        title: str | None = None,
        description: str | None = None,
        annotations: mcp.types.ToolAnnotations | None = None,
        structured_output: bool | None = None,
    ) -> Callable[[mcp.types.AnyFunction], mcp.types.AnyFunction]:
        """Decorator to register a tool.

        Tools can optionally request a Context object by adding a parameter with the
        Context type annotation. The context provides access to MCP capabilities like
        logging, progress reporting, and resource access.

        Args:
            name: Optional name for the tool (defaults to function name)
            title: Optional human-readable title for the tool
            description: Optional description of what the tool does
            annotations: Optional ToolAnnotations providing additional tool information
            structured_output: Controls whether the tool's output is structured or unstructured
                - If None, auto-detects based on the function's return type annotation
                - If True, unconditionally creates a structured tool (return type annotation permitting)
                - If False, unconditionally creates an unstructured tool

        Example:
            @server.tool()
            def my_tool(x: int) -> str:
                return str(x)

            @server.tool()
            def tool_with_context(x: int, ctx: Context) -> str:
                ctx.info(f"Processing {x}")
                return str(x)

            @server.tool()
            async def async_tool(x: int, context: Context) -> str:
                await context.report_progress(50, 100)
                return str(x)
        """
        # Check if user passed function directly instead of calling decorator
        if callable(name):
            raise TypeError(
                "The @tool decorator was used incorrectly. Did you forget to call it? Use @tool() instead of @tool"
            )

        def decorator(fn: mcp.types.AnyFunction) -> mcp.types.AnyFunction:
            self.add_tool(
                fn,
                name=name,
                title=title,
                description=description,
                annotations=annotations,
                structured_output=structured_output,
            )
            return fn

        return decorator

    def add_tool(
        self,
        fn: mcp.types.AnyFunction,
        name: str | None = None,
        title: str | None = None,
        description: str | None = None,
        annotations: mcp.types.ToolAnnotations | None = None,
        structured_output: bool | None = None,
    ) -> None:
        self._tools.append(
            Tool(
                fn=fn,
                name=name,
                title=title,
                description=description,
                annotations=annotations,
                structured_output=structured_output,
            )
        )

    def application(self) -> Application:
        """
        Returns a Reboot `Application` for running the MCP tools,
        resources, prompts, etc that were defined.
        """
        application = Application(servicers=self.servicers())

        application.http.mount(
            self._path,
            factory=self.streamable_http_app_factory,
        )

        return application

    @property
    def streamable_http_app_factory(self):
        return functools.partial(
            _streamable_http_app,
            self._log_level,
            self._path,
            self._resources,
            self._prompts,
            self._tools,
        )


def _streamable_http_app(
    log_level: LogLevel,
    path: str,
    resources: list[Resource],
    prompts: list[Prompt],
    tools: list[Tool],
    external_context_from_request: Callable[[Request], ExternalContext],
):
    mcp = fastmcp.FastMCP(log_level=log_level)

    for resource in resources:
        mcp.resource(
            resource.uri,
            name=resource.name,
            title=resource.title,
            description=resource.description,
            mime_type=resource.mime_type,
        )(resource.fn)

    for prompt in prompts:
        mcp.prompt(
            name=prompt.name,
            title=prompt.title,
            description=prompt.description,
        )(prompt.func)

    for tool in tools:
        mcp.add_tool(
            _wrap_tool(tool.fn),
            name=tool.name,
            title=tool.title,
            description=tool.description,
            annotations=tool.annotations,
            structured_output=tool.structured_output,
        )

    _servers[path] = mcp._mcp_server

    return Starlette(
        routes=[
            Route(
                "/{path:path}",
                endpoint=StreamableHTTPASGIApp(
                    path,
                    external_context_from_request,
                ),
            ),
        ],
    )


def _wrap_tool(fn: mcp.types.AnyFunction) -> mcp.types.AnyFunction:
    signature = inspect.signature(fn)

    wrapper_parameters = [
        # Always include the `context` parameter so we can access
        # session specific things.
        inspect.Parameter(
            "ctx",
            kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            annotation=fastmcp.Context,
        )
    ]

    context_parameter_names = []

    for parameter_name, parameter in signature.parameters.items():
        annotation = parameter.annotation
        if (
            isinstance(annotation, type) and
            issubclass(annotation, fastmcp.Context)
        ):
            raise TypeError(
                "`DurableMCP` only injects `DurableContext` not `Context`")
        if (
            isinstance(annotation, type) and
            issubclass(annotation, DurableContext)
        ):
            context_parameter_names.append(parameter_name)
        else:
            wrapper_parameters.append(parameter)

    wrapper_signature = signature.replace(parameters=wrapper_parameters)

    async def wrapper(ctx: fastmcp.Context, *args, **kwargs):

        context: WorkflowContext | None = _context.get()

        assert context is not None

        # To account for the lack of "intersection" types in
        # Python (which is actively being worked on), we instead
        # create a new dynamic `DurableContext` instance that
        # inherits from the instance of `WorkflowContext` that we
        # already have.
        context.__class__ = DurableContext

        context = cast(DurableContext, context)

        # Now we add the `DurableContextProtocol` properties.
        context._event_aliases = set()

        context.session = DurableSession(ctx.session, context)

        async def report_progress(
            self,
            progress: float,
            total: float | None = None,
            message: str | None = None,
        ) -> None:
            progress_token = (
                ctx.request_context.meta.progressToken
                if ctx.request_context.meta else None
            )

            if progress_token is None:
                return

            # TODO: consider tracking all reported progress and if
            # it has gone "backwards" provide a nice error so
            # developers can fix their bug (presumably they don't
            # ever want the progress to go "backwards").

            event_alias = (
                f"report_progress(progress={progress}, total={total}, "
                f"message={message})"
            )

            assert context is not None

            if context.within_loop():
                event_alias += f" #{context.task.iteration}"

            if event_alias in self._event_aliases:
                raise TypeError(
                    f"Looks like you're calling `report_progress()` "
                    "more than once with the same arguments"
                )

            self._event_aliases.add(event_alias)

            workflow_id = context.workflow_id

            assert workflow_id is not None

            # Generate a unique but deterministic ID for this
            # event based on the alias and this workflow (which is
            # unique per request).
            event_id = uuid5(workflow_id, event_alias).hex

            await ctx.session.send_notification(
                mcp.types.ServerNotification(
                    mcp.types.ProgressNotification(
                        # TODO: figure out why `mypy` requires
                        # passing `method` which has a default.
                        method="notifications/progress",
                        params=mcp.types.ProgressNotificationParams(
                            progressToken=progress_token,
                            progress=progress,
                            total=total,
                            message=message,
                            _meta=mcp.types.NotificationParams.Meta(
                                rebootEventId=str(event_id),
                            ),
                        ),
                    ),
                ),
                related_request_id=ctx.request_id,
            )

        context.report_progress = MethodType(report_progress, context)  # type: ignore[method-assign]

        async def log(
            self,
            level: Literal["debug", "info", "warning", "error"],
            message: str,
            *,
            logger_name: str | None = None,
        ) -> None:
            event_alias = (
                f"log(level='{level}', message='{message}', logger_name={logger_name})"
            )

            assert context is not None

            if context.within_loop():
                event_alias += f" #{context.task.iteration}"

            if event_alias in self._event_aliases:
                raise TypeError(
                    "Looks like you're trying to `log()` "
                    "more than once with the same arguments"
                )

            self._event_aliases.add(event_alias)

            workflow_id = context.workflow_id

            assert workflow_id is not None

            # Generate a unique but deterministic ID for this
            # event based on the alias and this workflow (which is
            # unique per request).
            event_id = uuid5(workflow_id, event_alias).hex

            await ctx.session.send_notification(
                mcp.types.ServerNotification(
                    mcp.types.LoggingMessageNotification(
                        # TODO: figure out why `mypy` requires
                        # passing `method` which has a default.
                        method="notifications/message",
                        params=mcp.types.LoggingMessageNotificationParams(
                            level=level,
                            data=message,
                            logger=logger_name,
                            _meta=mcp.types.NotificationParams.Meta(
                                rebootEventId=str(event_id),
                            ),
                        ),
                    )
                ),
                related_request_id=ctx.request_id,
            )

        context.log = MethodType(log, context)  # type: ignore[method-assign]

        async def debug(self, message: str) -> None:
            await self.log("debug", message)

        context.debug = MethodType(debug, context)  # type: ignore[method-assign]

        async def info(self, message: str) -> None:
            await self.log("info", message)

        context.info = MethodType(info, context)  # type: ignore[method-assign]

        async def warning(self, message: str) -> None:
            await self.log("warning", message)

        context.warning = MethodType(warning, context)  # type: ignore[method-assign]

        async def error(self, message: str) -> None:
            await self.log("error", message)

        context.error = MethodType(error, context)  # type: ignore[method-assign]

        async def elicit(
            self,
            message: str,
            schema: type[ElicitSchemaModelT],
        ) -> ElicitationResult:
            event_alias = (
                f"elicit(message='{message}', schema={type(schema).__name__})"
            )

            assert context is not None

            if context.within_loop():
                event_alias += f" #{context.task.iteration}"

            if event_alias in self._event_aliases:
                raise TypeError(
                    "Looks like you're trying to `elicit()` "
                    "more than once with the same arguments"
                )

            self._event_aliases.add(event_alias)

            workflow_id = context.workflow_id

            assert workflow_id is not None

            memoize = Memoize.ref(uuid5(workflow_id, event_alias).hex)

            # Initial reset, only done once per workflow, we've
            # already accounted for a possible loop iteration in
            # the `event_alias` above.
            await memoize.per_workflow(event_alias).Reset(context)

            status = await memoize.always().Status(context)

            if not status.started:
                await memoize.always().Start(context)
            else:
                message = (
                    f"Sorry, we got disconnected and need to try again: {message}"
                )

            async def send_request_and_wait_for_result():
                # Generate a unique and _random_ ID for this event because
                # we want to send it _everytime_ since the client is not
                # also durable.
                event_id = uuid4().hex

                # THIS CODE IS MORE OR LESS COPIED FROM
                # `mcp.server.elicitation.elicit_with_validation()`
                # because there was not a good way to override that
                # functionality.
                #
                # Validate that schema only contains primitive types and
                # fail loudly if not.
                _validate_elicitation_schema(schema)

                json_schema = schema.model_json_schema()

                return await ctx.session.send_request(
                    mcp.types.ServerRequest(
                        mcp.types.ElicitRequest(
                            method="elicitation/create",
                            params=mcp.types.ElicitRequestParams(
                                message=message,
                                requestedSchema=json_schema,
                                _meta=mcp.types.RequestParams.Meta(
                                    rebootEventId=str(event_id),
                                ),
                            ),
                        )
                    ),
                    mcp.types.ElicitResult,
                    metadata=ServerMessageMetadata(
                        related_request_id=ctx.request_id,
                    ),
                )

            result = await at_least_once(
                "Send request, wait for result",
                context,
                send_request_and_wait_for_result,
                type=mcp.types.ElicitResult,
            )

            if result.action == "accept" and result.content:
                # Validate and parse the content using the schema.
                validated_data = schema.model_validate(result.content)
                return AcceptedElicitation(data=validated_data)
            elif result.action == "decline":
                return DeclinedElicitation()
            elif result.action == "cancel":
                return CancelledElicitation()
            else:
                # This should never happen, but handle it just in case.
                raise ValueError(
                    f"Unexpected elicitation action: {result.action}"
                )

        context.elicit = MethodType(elicit, context)  # type: ignore[method-assign]

        for context_parameter_name in context_parameter_names:
            kwargs[context_parameter_name] = context

        bound = signature.bind(*args, **kwargs)
        bound.apply_defaults()

        try:
            if fastmcp.tools.base._is_async_callable(fn):
                return await fn(**dict(bound.arguments))

            return fn(**dict(bound.arguments))
        except:
            # TODO: print stack trace after we've fixed `memoize`
            # effect validation bug.
            #
            # import traceback
            # traceback.print_exc()
            raise

    setattr(wrapper, "__signature__", wrapper_signature)
    wrapper.__name__ = fn.__name__
    wrapper.__doc__ = fn.__doc__

    return wrapper



class StreamableHTTPASGIApp:
    """
    ASGI application for Streamable HTTP server transport.
    """

    def __init__(
        self,
        path: str,
        external_context_from_request: Callable[[Request], ExternalContext],
    ):
        self._path = path
        self._external_context_from_request = external_context_from_request
        self._http_transports: dict[str, StreamableHTTPServerTransport] = {}
        self._connect_tasks: dict[str, asyncio.Task] = {}

    async def __call__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
    ) -> None:
        request = Request(scope, receive)

        if request.headers.get(STATE_REF_HEADER) is None:
            mcp_session_id = request.headers.get(MCP_SESSION_ID_HEADER)

            # If this is a new session, i.e., we don't have an ID,
            # then generate one which we'll use as the session ID.
            mcp_session_id = mcp_session_id or uuid4().hex

            headers = dict(request.headers)
            headers[STATE_REF_HEADER] = f"rbt.mcp.v1.Session:{mcp_session_id}"

            # Need to delete consensus ID header so that we'll
            # properly get routed.
            del headers[CONSENSUS_ID_HEADER]

            response: StreamingResponse | Response

            try:
                async with httpx.AsyncClient() as client:
                    # Too simplify we always perform a streaming
                    # request even if the response is not streaming.
                    async with client.stream(
                        request.method,
                        str(request.url),
                        headers=headers,
                        content=await request.body(),
                    ) as upstream:
                        # Create a generator to yield chunks from the
                        # upstream response.
                        async def streamer():
                            async for chunk in upstream.aiter_bytes():
                                yield chunk

                        response = StreamingResponse(
                            content=streamer(),
                            status_code=upstream.status_code,
                            headers=upstream.headers,
                        )
                        await response(scope, receive, send)
                        return
            except Exception as e:
                response = Response(
                    f"Proxy request failed: {e}",
                    status_code=500,
                )
                await response(scope, receive, send)
                return

        # This request has properly been forwarded to the consensus
        # responsible for this session.
        context = self._external_context_from_request(request)

        mcp_session_id = request.headers.get(MCP_SESSION_ID_HEADER)

        if (
            mcp_session_id is not None and
            mcp_session_id in self._http_transports
        ):
            transport = self._http_transports[mcp_session_id]
            await transport.handle_request(scope, receive, send)
            return

        # If this is a new session, i.e., `mcp_session_id is None`, we
        # need to use the session ID we already generated so all
        # requests for it will get routed to this consensus.
        if mcp_session_id is None:
            session_ref = request.headers.get(STATE_REF_HEADER)
            assert session_ref is not None
            session_id = StateRef.from_maybe_readable(session_ref).id
            mcp_session_id = session_id

        http_transport = StreamableHTTPServerTransport(
            mcp_session_id=mcp_session_id,
            is_json_response_enabled=False,
            event_store=DurableEventStore(context),
            security_settings=None,
        )

        session = Session.ref(mcp_session_id)

        self._http_transports[mcp_session_id] = http_transport

        started = asyncio.Event()

        async def connect():

            async with http_transport.connect() as streams:
                started.set()
                read_stream, write_stream = streams

                writer_tasks: list[asyncio.Task] = []

                async def reader():
                    async for message in read_stream:
                        # TODO: we can't pickle `request_context`
                        # which is a `starlette.requests.Request`,
                        # but it looks like it is only for
                        # _advanced_ use cases, not really
                        # documented
                        # (https://github.com/modelcontextprotocol/python-sdk/blob/main/README.md?plain=1#L948),
                        # and if we really want we could probably
                        # grab what is necessary and send if along
                        # in a picklable way.
                        message.metadata.request_context = None  # type: ignore
                        # TODO: ideally we spawn `HandleMessage`
                        # _before_ a 202 Accepted is sent.
                        await session.spawn().HandleMessage(
                            context,
                            path=self._path,
                            message_bytes=pickle.dumps(message),
                        )

                        if not isinstance(message, Exception) and isinstance(
                            message.message.root,
                            mcp.types.JSONRPCRequest,
                        ):
                            request_id = message.message.root.id

                            async def writer():
                                async for message, _ in replay(
                                    context,
                                    stream_id=str(request_id),
                                ):
                                    await write_stream.send(message)

                            writer_tasks.append(asyncio.create_task(writer()))

                reader_task = asyncio.create_task(reader())

                try:
                    # Await the reader first, since it will be
                    # finished once the session is finished, then
                    # we can cancel any writers.
                    await reader_task
                finally:
                    reader_task.cancel()
                    for writer_task in writer_tasks:
                        writer_task.cancel()
                    await asyncio.wait(
                        [reader_task] + writer_tasks,
                        return_when=asyncio.ALL_COMPLETED,
                    )

        self._connect_tasks[mcp_session_id] = asyncio.create_task(connect())

        def done(task):
            try:
                task.result()
            except asyncio.CancelledError:
                pass
            except Exception as e:
                logger.warning(f"Session {mcp_session_id} crashed: {e}")
            finally:
                if http_transport.is_terminated:
                    assert mcp_session_id in self._connect_tasks
                    del self._connect_tasks[mcp_session_id]
                    assert mcp_session_id in self._http_transports
                    del self._http_transports[mcp_session_id]

        self._connect_tasks[mcp_session_id].add_done_callback(done)

        await started.wait()

        await http_transport.handle_request(scope, receive, send)



# yapf: disable
# isort: skip_file
# ruff: noqa



# To not generate code where imported names might get shadowed when a user
# specifies some name in their proto file to be the same as one of our imported
# names, (for example: a request field named `uuid`) we bind all imports to
# names that are forbidden in 'proto' and therefore can never collide.

# Standard imports.
from __future__ import annotations as IMPORT_future_annotations

# The following MUST appear before the rest of the imports, since those imports
# may be invalid (broken) if the generated code is mismatched with the installed
# libraries.
import rebootdev.versioning as IMPORT_reboot_versioning
IMPORT_reboot_versioning.check_generated_code_compatible("0.37.1")

# ATTENTION: no types in this file should be imported with their unqualified
#            name (e.g. `from typing import Any`). That would cause clashes
#            with user-defined methods that have the same name. Use
#            fully-qualified names (e.g. `IMPORT_typing.Any`) instead.
import asyncio as IMPORT_asyncio
import builtins as IMPORT_builtins
import contextvars as IMPORT_contextvars
import dataclasses as IMPORT_dataclasses
import google.protobuf.descriptor as IMPORT_google_protobuf_descriptor
import google.protobuf.json_format as IMPORT_google_protobuf_json_format
import google.protobuf.message as IMPORT_google_protobuf_message
import grpc as IMPORT_grpc
import grpc_status._async as IMPORT_rpc_status_async
from grpc_status import rpc_status as IMPORT_rpc_status_sync
import json as IMPORT_json
import os as IMPORT_os
import traceback as IMPORT_traceback
import uuid as IMPORT_uuid
import pickle as IMPORT_pickle
import rebootdev as IMPORT_rebootdev
import log.log as IMPORT_log_log   # type: ignore[import]
import typing as IMPORT_typing
import rebootdev.aio.backoff as IMPORT_reboot_aio_backoff
import functools as IMPORT_functools
from abc import abstractmethod as IMPORT_abc_abstractmethod
from datetime import datetime as IMPORT_datetime_datetime
from datetime import timedelta as IMPORT_datetime_timedelta
from datetime import timezone as IMPORT_datetime_timezone
from google.protobuf import timestamp_pb2 as IMPORT_google_protobuf_timestamp_pb2
from google.protobuf import wrappers_pb2 as IMPORT_google_protobuf_wrappers_pb2
import rebootdev.aio.tracing as IMPORT_reboot_aio_tracing
from google.rpc import status_pb2 as IMPORT_google_rpc_status_pb2
from tzlocal import get_localzone as IMPORT_tzlocal_get_localzone
import rebootdev.aio.call as IMPORT_reboot_aio_call
import rebootdev.aio.contexts as IMPORT_reboot_aio_contexts
import rebootdev.aio.headers as IMPORT_reboot_aio_headers
import rebootdev.aio.idempotency as IMPORT_reboot_aio_idempotency
import rebootdev.aio.internals.channel_manager as IMPORT_reboot_aio_internals_channel_manager
import rebootdev.aio.internals.middleware as IMPORT_reboot_aio_internals_middleware
import rebootdev.aio.internals.tasks_cache as IMPORT_reboot_aio_internals_tasks_cache
import rebootdev.aio.internals.tasks_dispatcher as IMPORT_reboot_aio_internals_tasks_dispatcher
import rebootdev.aio.placement as IMPORT_reboot_aio_placement
import rebootdev.aio.servicers as IMPORT_reboot_aio_servicers
import rebootdev.aio.state_managers as IMPORT_reboot_aio_state_managers
import rebootdev.aio.stubs as IMPORT_reboot_aio_stubs
import rebootdev.aio.tasks as IMPORT_reboot_aio_tasks
import rebootdev.aio.types as IMPORT_reboot_aio_types
import rebootdev.aio.external as IMPORT_reboot_aio_external
import rebootdev.aio.workflows as IMPORT_reboot_aio_workflows
import rebootdev.settings as IMPORT_reboot_settings
import rebootdev.nodejs.python as IMPORT_reboot_nodejs_python
from rebootdev.time import DateTimeWithTimeZone as IMPORT_reboot_time_DateTimeWithTimeZone
import rbt.v1alpha1 as IMPORT_rbt_v1alpha1
import rbt.v1alpha1.nodejs_pb2 as IMPORT_rbt_v1alpha1_nodejs_pb2
import google.protobuf.any_pb2 as IMPORT_google_protobuf_any_pb2

# User defined or referenced imports.
import google.protobuf.descriptor_pb2
# Additionally re-export all messages and enums from the pb2 module.
from google.protobuf.descriptor_pb2 import (
    DescriptorProto,
    Edition,
    EnumDescriptorProto,
    EnumOptions,
    EnumValueDescriptorProto,
    EnumValueOptions,
    ExtensionRangeOptions,
    FeatureSet,
    FeatureSetDefaults,
    FieldDescriptorProto,
    FieldOptions,
    FileDescriptorProto,
    FileDescriptorSet,
    FileOptions,
    GeneratedCodeInfo,
    MessageOptions,
    MethodDescriptorProto,
    MethodOptions,
    OneofDescriptorProto,
    OneofOptions,
    ServiceDescriptorProto,
    ServiceOptions,
    SourceCodeInfo,
    UninterpretedOption,
)

logger = IMPORT_log_log.get_logger(__name__)

############################ Legacy gRPC Servicers ############################
# This section is relevant (only) for servicers that implement a legacy gRPC
# service in a Reboot context. It is irrelevant to clients.

def MakeLegacyGrpcServiceable(
    # A legacy gRPC servicer type can't be more specific than `type`,
    # because legacy gRPC servicers (as generated by the gRPC `protoc`
    # plugin) do not share any common base class other than `object`.
    servicer_type: type
) -> IMPORT_reboot_aio_servicers.Serviceable:
    raise ValueError(f"Unknown legacy gRPC servicer type '{servicer_type}'")



############################ Reboot Servicer Middlewares ############################
# This section is relevant (only) for servicers implementing a Reboot servicer. It
# is irrelevant to clients, except for the fact that some clients are _also_ such
# servicers.

# For internal calls, we can use a magic token to bypass token verification and
# authorization checks. The token provides no auth information (e.g.,
# `context.auth is None`).
__internal_magic_token__: str = f'internal-{str(IMPORT_uuid.uuid4())}'


############################ Client Stubs ############################
# This section is relevant for clients accessing a Reboot service. Since
# servicers are themselves often clients also, this code is generated for
# them also.



############################ Authorizers ############################
# Relevant to servicers; irrelevant to clients.


############################ Reboot Servicers ############################
# Base classes for server-side implementations of Reboot servicers.
# Irrelevant to clients.


############################ Clients ############################
# The main developer-facing entrypoints for any Reboot type. Relevant to both
# clients and servicers (who use it to find the right servicer base types, as well
# as often being clients themselves).

# Attach an explicit time time zone to "naive" `datetime` objects. A "naive" `datetime` doesn't have a
# time zone. Such objects are typically interpreted as representing local time, but could be confused
# for objects representing UTC. This helper function disambiguates by explicitly attaching the local
# time zone to `datetime` objects that don't already have an explicit time zone. If the `datetime` object
# is already timezone-aware, we still convert it to our custom `DateTimeWithTimeZone` type.
def ensure_has_timezone(
    *,
    when: IMPORT_typing.Optional[IMPORT_datetime_datetime | IMPORT_datetime_timedelta] = None,
) -> IMPORT_typing.Optional[IMPORT_reboot_time_DateTimeWithTimeZone | IMPORT_datetime_timedelta]:
    if isinstance(when, IMPORT_datetime_datetime):
        return IMPORT_reboot_time_DateTimeWithTimeZone.from_datetime(when)
    return when


############################ Servicer Node adapters ############################
# Used by Node.js servicer implementations to access Python code and vice-versa.
# Relevant to servicers, irrelevant to clients.


############################ Reference Node adapters ############################
# Used by Node.js WeakReference implementations to access Python code and
# vice-versa. Relevant to clients.

# yapf: enable



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rbt/mcp/v1/session.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from rbt.v1alpha1 import options_pb2 as rbt_dot_v1alpha1_dot_options__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18rbt/mcp/v1/session.proto\x12\nrbt.mcp.v1\x1a\x1arbt/v1alpha1/options.proto\"\x0f\n\x07Session:\x04\x82\xb5\x18\x00\";\n\x14HandleMessageRequest\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x15\n\rmessage_bytes\x18\x02 \x01(\x0c\"\x17\n\x15HandleMessageResponse\"1\n\nRunRequest\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x15\n\rmessage_bytes\x18\x02 \x01(\x0c\"\r\n\x0bRunResponse2\xae\x01\n\x0eSessionMethods\x12\\\n\rHandleMessage\x12 .rbt.mcp.v1.HandleMessageRequest\x1a!.rbt.mcp.v1.HandleMessageResponse\"\x06\x82\xb5\x18\x02\"\x00\x12>\n\x03Run\x12\x16.rbt.mcp.v1.RunRequest\x1a\x17.rbt.mcp.v1.RunResponse\"\x06\x82\xb5\x18\x02\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'rbt.mcp.v1.session_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_SESSION']._loaded_options = None
  _globals['_SESSION']._serialized_options = b'\202\265\030\000'
  _globals['_SESSIONMETHODS'].methods_by_name['HandleMessage']._loaded_options = None
  _globals['_SESSIONMETHODS'].methods_by_name['HandleMessage']._serialized_options = b'\202\265\030\002\"\000'
  _globals['_SESSIONMETHODS'].methods_by_name['Run']._loaded_options = None
  _globals['_SESSIONMETHODS'].methods_by_name['Run']._serialized_options = b'\202\265\030\002\"\000'
  _globals['_SESSION']._serialized_start=68
  _globals['_SESSION']._serialized_end=83
  _globals['_HANDLEMESSAGEREQUEST']._serialized_start=85
  _globals['_HANDLEMESSAGEREQUEST']._serialized_end=144
  _globals['_HANDLEMESSAGERESPONSE']._serialized_start=146
  _globals['_HANDLEMESSAGERESPONSE']._serialized_end=169
  _globals['_RUNREQUEST']._serialized_start=171
  _globals['_RUNREQUEST']._serialized_end=220
  _globals['_RUNRESPONSE']._serialized_start=222
  _globals['_RUNRESPONSE']._serialized_end=235
  _globals['_SESSIONMETHODS']._serialized_start=238
  _globals['_SESSIONMETHODS']._serialized_end=412
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rbt/mcp/v1/stream.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from rbt.v1alpha1 import options_pb2 as rbt_dot_v1alpha1_dot_options__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17rbt/mcp/v1/stream.proto\x12\nrbt.mcp.v1\x1a\x1bgoogle/protobuf/empty.proto\x1a\x1arbt/v1alpha1/options.proto\"1\n\x06Stream\x12!\n\x06\x65vents\x18\x01 \x03(\x0b\x32\x11.rbt.mcp.v1.Event:\x04\x82\xb5\x18\x00\"*\n\x05\x45vent\x12\n\n\x02id\x18\x01 \x01(\t\x12\x15\n\rmessage_bytes\x18\x02 \x01(\x0c\"5\n\nPutRequest\x12\x10\n\x08\x65vent_id\x18\x01 \x01(\t\x12\x15\n\rmessage_bytes\x18\x02 \x01(\x0c\"\r\n\x0bPutResponse\"=\n\rReplayRequest\x12\x1a\n\rlast_event_id\x18\x01 \x01(\tH\x00\x88\x01\x01\x42\x10\n\x0e_last_event_id\"3\n\x0eReplayResponse\x12!\n\x06\x65vents\x18\x01 \x03(\x0b\x32\x11.rbt.mcp.v1.Event2\xda\x01\n\rStreamMethods\x12@\n\x06\x43reate\x12\x16.google.protobuf.Empty\x1a\x16.google.protobuf.Empty\"\x06\x82\xb5\x18\x02\x12\x00\x12>\n\x03Put\x12\x16.rbt.mcp.v1.PutRequest\x1a\x17.rbt.mcp.v1.PutResponse\"\x06\x82\xb5\x18\x02\x12\x00\x12G\n\x06Replay\x12\x19.rbt.mcp.v1.ReplayRequest\x1a\x1a.rbt.mcp.v1.ReplayResponse\"\x06\x82\xb5\x18\x02\n\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'rbt.mcp.v1.stream_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_STREAM']._loaded_options = None
  _globals['_STREAM']._serialized_options = b'\202\265\030\000'
  _globals['_STREAMMETHODS'].methods_by_name['Create']._loaded_options = None
  _globals['_STREAMMETHODS'].methods_by_name['Create']._serialized_options = b'\202\265\030\002\022\000'
  _globals['_STREAMMETHODS'].methods_by_name['Put']._loaded_options = None
  _globals['_STREAMMETHODS'].methods_by_name['Put']._serialized_options = b'\202\265\030\002\022\000'
  _globals['_STREAMMETHODS'].methods_by_name['Replay']._loaded_options = None
  _globals['_STREAMMETHODS'].methods_by_name['Replay']._serialized_options = b'\202\265\030\002\n\000'
  _globals['_STREAM']._serialized_start=96
  _globals['_STREAM']._serialized_end=145
  _globals['_EVENT']._serialized_start=147
  _globals['_EVENT']._serialized_end=189
  _globals['_PUTREQUEST']._serialized_start=191
  _globals['_PUTREQUEST']._serialized_end=244
  _globals['_PUTRESPONSE']._serialized_start=246
  _globals['_PUTRESPONSE']._serialized_end=259
  _globals['_REPLAYREQUEST']._serialized_start=261
  _globals['_REPLAYREQUEST']._serialized_end=322
  _globals['_REPLAYRESPONSE']._serialized_start=324
  _globals['_REPLAYRESPONSE']._serialized_end=375
  _globals['_STREAMMETHODS']._serialized_start=378
  _globals['_STREAMMETHODS']._serialized_end=596
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class Session(google.protobuf.message.Message):
    """//////////////////////////////////////////////////////////////////////"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___Session = Session

@typing_extensions.final
class HandleMessageRequest(google.protobuf.message.Message):
    """//////////////////////////////////////////////////////////////////////"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PATH_FIELD_NUMBER: builtins.int
    MESSAGE_BYTES_FIELD_NUMBER: builtins.int
    path: builtins.str
    message_bytes: builtins.bytes
    def __init__(
        self,
        *,
        path: builtins.str = ...,
        message_bytes: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["message_bytes", b"message_bytes", "path", b"path"]) -> None: ...

global___HandleMessageRequest = HandleMessageRequest

@typing_extensions.final
class HandleMessageResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___HandleMessageResponse = HandleMessageResponse

@typing_extensions.final
class RunRequest(google.protobuf.message.Message):
    """//////////////////////////////////////////////////////////////////////"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PATH_FIELD_NUMBER: builtins.int
    MESSAGE_BYTES_FIELD_NUMBER: builtins.int
    path: builtins.str
    message_bytes: builtins.bytes
    def __init__(
        self,
        *,
        path: builtins.str = ...,
        message_bytes: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["message_bytes", b"message_bytes", "path", b"path"]) -> None: ...

global___RunRequest = RunRequest

@typing_extensions.final
class RunResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___RunResponse = RunResponse
